#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <map>
#include <set>

typedef struct tnode *Tptr;
typedef struct tnode {
   char splitchar;
   unsigned int hash;
   unsigned int no;
   Tptr lokid, eqkid, hikid;
} Tnode;


int rsearch(Tptr p, char *s)

{   if (!p) return 0;
    if (*s < p->splitchar)
         return rsearch(p->lokid, s);
    else if (*s > p->splitchar)
        return rsearch(p->hikid, s);
    else {
        if (*s == 0) return 1;
        return rsearch(p->eqkid, ++s);
   }
}


Tptr insert(Tptr p, char *s)
{ if (p == 0) {
       p = (Tptr) malloc(sizeof(Tnode));
       p->splitchar = *s;
       p->lokid = p->eqkid = p->hikid = 0;
       p->no = 0;
  }
  if (*s < p->splitchar)
       p->lokid = insert(p->lokid, s);
  else if (*s == p->splitchar) {
       if (*s != 0)
           p->eqkid = insert(p->eqkid, ++s);
  } else
       p->hikid = insert(p->hikid, s);
  return p;
}


Tptr rinsall(Tptr root, char *a[], int n)
{   int m;
    if (n < 1) return root;
    m = n/2;
    root = insert(root,a[m]);
    rinsall(root, a, m);
    rinsall(root, a + m+1, n-m-1);
    return root;
}

static unsigned int calculate_hash(Tptr node)
{
    if(node == NULL)
        return 0;

    // Calculate the hashes for children recursively and use them for the current hash
    node->hash = (node->splitchar - 'a') + 31 * calculate_hash(node->eqkid);
    node->hash ^= calculate_hash(node->lokid);
    node->hash ^= calculate_hash(node->hikid);
    node->hash ^= (node->hash >> 16);

    return node->hash;
}


static bool equal(Tptr h1, Tptr h2)
{
    if(h1 == NULL || h2 == NULL)
        return h1 == NULL && h2 == NULL;

    return h1->splitchar == h2->splitchar &&
           equal(h1->lokid, h2->lokid) &&
           equal(h1->eqkid, h2->eqkid) &&
           equal(h1->hikid, h2->hikid);
}

static std::map<unsigned int, std::vector<Tptr> > hashmap;

static bool check_and_remove_duplicate(Tptr& node)
{
    std::vector<Tptr>& v = hashmap[node->hash];

    for (size_t i = 0; i < v.size(); ++i)
    {
        if (equal(v[i], node))
        {
            node = v[i];
            return false;
        }
    }

    v.push_back(node);

    return true;
}

static void remove_duplicates(Tptr node)
{
    if (node->lokid)
        if(check_and_remove_duplicate(node->lokid))
            remove_duplicates(node->lokid);

    if (node->eqkid)
        if (check_and_remove_duplicate(node->eqkid))
            remove_duplicates(node->eqkid);

    if (node->hikid)
        if (check_and_remove_duplicate(node->hikid))
            remove_duplicates(node->hikid);
}


static std::vector<Tptr> allnodes;
static void populate(Tptr node)
{
    if (node == NULL)
        return;

    if (node->no == 0)
    {
        allnodes.push_back(node);
        node->no = allnodes.size();
    }

    populate(node->lokid);
    populate(node->eqkid);
    populate(node->hikid);
}

int main(int argc, char* argv[])
{
    if (argc != 3)
    {
        printf("usage: buildtst input output\n");
        return 0;
    }

    std::vector<char*> words;

    FILE* fis = fopen(argv[1], "rt");
    while (1)
    {
        char buf[256];
        if (fgets(buf, 256, fis) == NULL)
            break;
        int s = strlen(buf);
        if (s && buf[s - 1] == 0xa)
            buf[--s] = 0;
        if (s && buf[s - 1] == 0xd)
            buf[--s] = 0;
        if (s)
            words.push_back(strdup(buf));
    }
    fclose(fis);

    Tptr root = rinsall(NULL, &words[0], words.size());

    calculate_hash(root);
    remove_duplicates(root);
    populate(root);

    FILE* fos = fopen(argv[2], "wb");
    for (size_t i = 0; i < allnodes.size(); ++i)
    {
        Tptr node = allnodes[i];

        fwrite(&node->splitchar, 1, 1, fos);

        int lono = node->lokid ? node->lokid->no : 0;
        int eqno = node->eqkid ? node->eqkid->no : 0;
        int hino = node->hikid ? node->hikid->no : 0;

        fwrite(&lono, 3, 1, fos);
        fwrite(&eqno, 3, 1, fos);
        fwrite(&hino, 3, 1, fos);
    }
    fclose(fos);

/*    printf("%d\n", allnodes.size());

    for (size_t i = 0; i < words.size(); ++i)
        if (rsearch(root, words[i]) == 0)
            printf("this shouldn't happen!\n");
*/
    return 0;
}





